arXiv is now an independent nonprofit! Learn more
License: CC BY-NC-SA 4.0
arXiv:2608.05033v3 [cs.DC] 10 Sep 2026

[Uncaptioned image]  SparseDitto: An Agentic Sparse Compilation Framework through Architecture-Aware Synthesis on GPUs

Conference: Anonymous Submission; 2026;
Shiyang Li email: li004074@umn.edu Affiliation: University of Minnesota , Minneapolis, MN, USA , Guangyan Sun email: sun01158@umn.edu Affiliation: University of Minnesota , Minneapolis, MN, USA , Jinwei Tang email: tang0940@umn.edu Affiliation: University of Minnesota , Minneapolis, MN, USA , Yanzhi Wang email: yanz.wang@northeastern.edu Affiliation: Northeastern University, Boston, MA, USA , Mingyi Hong email: mhong@umn.edu Affiliation: University of Minnesota , Minneapolis, MN, USA and Caiwen Ding email: dingc@umn.edu Affiliation: University of Minnesota , Minneapolis, MN, USA
2026
Abstract.

Sparse matrix computation performance on GPU depends on how representation and execution schedule match the input structure and target hardware. No single implementation consistently dominates across sparsity patterns, operators, and hardwares. Existing sparse compilers and specialized systems cannot cover all of them simultaneously.

We present SparseDitto, an agentic sparse compilation framework for sparse matrix computation on GPUs. It jointly synthesizes representation, execution schedule, and hardware mapping in a unified compilation plan. Structural analysis and a learned template-ranking prior guide architecture-aware synthesis. LLM-guided lowering realizes each plan as CUDA code, while target-GPU profiling drives plan refinement. SparseDitto covers multiple operators, e.g., SpMV, SpMM, and SpGEMM, and various representations within one framework. It can also automatically adapt to different hardwares. Across various SuiteSparse matrices, SparseDitto achieves geometric-mean speedups over cuSPARSE of 2.68×2.68\times on an NVIDIA RTX PRO 6000 and 2.79×2.79\times on an NVIDIA H200 (up to 146.61×\times). Its generated SpMM kernels accelerate full-batch GCN training by up to 3.39×3.39\times.

Keywords: 
sparse matrix computation, GPU kernel synthesis, sparse compilation framework

1. Introduction

Sparse matrix operations are core building blocks in scientific computing, graph analytics, and machine learning (Vázquez et al., 2010; Wang et al., 2023; Kjolstad et al., 2017; Liu and Vinter, 2015; Peng et al., 2024). Their performance on GPUs can change by orders of magnitude even when the matrix, operator, library, and GPU remain unchanged. For one matrix with 355 K nonzeros, cuSPARSE completes SpMM in 37 μ\mus using CSR but requires 13 ms using Blocked-ELL. Blocked-ELL regularizes sparse computation by storing each block row with a common number of blocks, which can expose efficient blocked execution. On this irregular matrix, however, the required padding makes much of the transferred data useless and removes that advantage. The 350×350\times performance difference is caused only by the representation and its associated kernel. This example exposes a basic challenge for sparse matrix computing on GPU. A high-performance sparse implementation needs to match its representation and execution strategy to the structure of the input.

Further fundamental challenge is that sparse GPU performance depends on three coupled implementation choices (Figure 1a). The sparse representation determines the layout of values and metadata. The execution schedule determines workload partitioning, thread mapping, and data movement. Hardware mapping further binds these choices to tile sizes, launch configurations, and on-chip resources on the target GPU.

Current state-of-the-art (SOTA) approaches follow three main directions. First, specialized sparse libraries and kernels encode expert-designed optimizations for particular operators, sparsity patterns, or hardware features (Cong et al., 2025; Fan et al., 2024; Wu et al., 2025). Second, sparse compilers and design-space search systems expose broader choices of representations and schedules (Kjolstad et al., 2017; Ye et al., 2023; Du et al., 2022). Third, using LLM-based CUDA programming agents to generate and optimize GPU kernels through iterative coding, compilation, profiling, and refinement (Zhu et al., 2025; Ouyang et al., 2025; Li et al., 2026). These directions provide optimization mechanisms, but they partially addressed the challenges. The key missing capability is to coordinate sparse-specific representation and scheduling decisions with flexible code construction and target-specific hardware adaptation.

We study three representative operator-specialized systems and one compiler-based system: CB-SpMV (Cong et al., 2025), DTC-SpMM (Fan et al., 2024), HSMU-SpGEMM (Wu et al., 2025), and SparseTIR (Ye et al., 2023). We obtain two key observations. First, sparse optimization is highly asymmetric. A compatible representation and schedule can provide meaningful gains, while a poor match can cause orders-of-magnitude slowdowns. The best alternative format improves SpMV by up to 1.47×1.47\times and SpMM by up to 2.8×2.8\times, while a poor pairing loses up to 350×350\times. Second, no fixed implementation consistently dominates across sparsity patterns, operators, and GPUs. CB-SpMV works well on cache-friendly blocked inputs. DTC-SpMM benefits matrices that keep enough useful work inside compressed Tensor-Core blocks. HSMU-SpGEMM performs well when its hash accumulator matches the intermediate-product distribution. Their relative performance also changes with the GPU architecture and CUDA runtime version.

(a) Design space of sparse matrix operators

Refer to caption

(b) SparseDitto workflow

Figure 1. SparseDitto’s design space and compilation workflow. Prior-work annotations in (a) highlight representative mechanisms.Two vertically stacked panels. The first shows the sparse kernel design space and SparseDitto's joint synthesis across representation, execution schedule, and hardware mapping. The second shows its compilation workflow, from a sparse workload and target GPU to a specialized kernel, with profile-grounded refinement.

The three SOTA directions leave complementary gaps. Specialized kernels cover only the sparsity patterns, operators, and hardware assumptions for which they were designed. Sparse compilers and design-space search systems broaden the optimization space, but their coverage is still bounded by developer-provided static rules. Search can also be expensive. AlphaSparse (Du et al., 2022), for example, can spend hours constructing a specialized SpMV implementation through offline search. General CUDA programming agents offer more flexible code construction, but they lack sparse-specific guidance for choosing representations and execution schedules. A functionally correct kernel can therefore remain inefficient when these high-level decisions do not match the input structure. A general solution needs both structured sparse optimization and the ability to adapt to the target GPU.

Our target is to end the game of designing a separate sparse optimization for every limited use case and hardware platform. We seek a framework that can combine existing sparse optimization strategies, use each where it fits, and adapt to new GPU platforms with minimal additional human effort. This goal leads to three requirements. The framework must jointly reason about representation, execution schedule, and hardware mapping. It must use structural analysis to guide these choices. It must also incorporate target-GPU measurements when static analysis is insufficient.

In this paper, we present SparseDitto, the first agentic sparse compilation framework that automatically specializes GPU kernels for a given sparse matrix, operator, and target GPU. As shown in Figure 1b, SparseDitto treats sparse kernel construction as a compilation problem. Architecture-aware synthesis builds a unified plan for representation, execution schedule, and hardware mapping. Agentic lowering realizes the plan as a specialized CUDA implementation for repeated execution. Subsequent calls reuse it for the same sparsity structure and operator configuration on the target GPU.

In summary, this paper makes the following contributions:

  • A unified sparse kernel compilation framework on GPUs. We formulate sparse GPU kernel compilation as the joint choice of data representation, execution schedule, and target-specific hardware mapping. The formulation captures the major implementation decisions shared by SpMV, SpMM, and SpGEMM within one framework.

  • Structure-guided and architecture-aware synthesis. A structural analysis computes 36 features that describe the sparsity pattern, representation overhead, and operator-dependent intermediate work. An interpretable additive energy model serves as a learned cost model and ranks established optimization templates from an offline measurement corpus. An architecture-aware synthesis stage combines this ranking with static workload bounds and target-GPU constraints to construct candidate compilation plans.

  • LLM-guided lowering with profile-guided optimization. SparseDitto lowers each compilation plan into CUDA using an LLM-guided coding stage and a set of supported sparse optimization mechanisms. Compilation and numerical validation check each candidate. Runtime measurements and profiling on the target GPU guide later changes to the representation, schedule, and hardware mapping. The fastest valid implementation is retained for reuse.

We evaluate SparseDitto on 6060 SuiteSparse matrices, covering SpMV, SpMM at four dense widths, and SpGEMM. SparseDitto reaches a geometric-mean speedup of 2.68×2.68\times over cuSPARSE on an RTX PRO 6000, with a maximum of 146.61×146.61\times. On an H200, it reaches 2.79×2.79\times, with a maximum of 78.5×78.5\times. The specialized SpMM kernels also accelerate full-batch GCN training by up to 3.39×3.39\times.

2. Background and Motivation

Figure 2. One 16×1616\times 16 matrix exposes intrinsic pattern cues and format-induced costs. The matrix contains an irregular dense tile, a high-degree column, an empty row, and uneven row lengths. CSR and CSC traverse compressed rows and columns, respectively. SELL-C with C=8C=8 introduces slice-local padding, while 4×44\times 4 BSR introduces block fill. The right column shows the resulting work granularity and imbalance for rows, columns, slice slots, and blocks.

2.1. Sparse Matrix Operators on GPUs

The sparsity pattern determines the distribution of nonzeros, the memory a kernel traverses, and the workload assigned to computing units. It also controls locality once the schedule is mapped to the GPU. The operator determines the work produced by each visited nonzero. As a result, the critical design choice changes across SpMV, SpMM, and SpGEMM.

SpMV computes y=Axy=Ax and performs only two floating-point operations per nonzero. Its performance often depends on sparse metadata traffic and the locality of gathers from xx. Load imbalance among parallel units is another major cost (Filippone et al., 2017; Liu and Vinter, 2015; Merrill and Garland, 2016). Row-based CSR avoids output reductions but inherits the row-length distribution. ELL and SELL use padding to regularize sparse traversal.

SpMM computes C=ABC=AB with a dense matrix BB of width KK. Each nonzero of AA drives KK multiply–adds, which creates another pattern dimension and more reuse of BB. The sparse representation needs to be coordinated with the mapping of rows or blocks to the GPU. The KK-tile and output-accumulator placement also affect the dataflow. Dense sparse tiles can enable Tensor Cores when useful arithmetic outweighs compaction and fill-in costs (Wang et al., 2023; Fan et al., 2024).

SpGEMM further discovers a sparse output. It computes C=ABC=AB for two sparse matrices inputs AA and BB. In a row-wise algorithm, row ii generates

(1) uB(i)=kA[i,:]nnz(B[k,:]),u_{B}(i)=\sum_{k\in A[i,:]}nnz(B[k,:]),

The distribution of uB(i)u_{B}(i) affects row partitioning and temporary storage. It also determines the suitability of dense, hash, merge, or hybrid accumulators (Srivastava et al., 2020; Nagasaka et al., 2017; Wu et al., 2025). Design space now extends beyond input traversal because the kernel must construct output indices and intermediate state.

Thus, the same pattern can produce different metadata streams, processing units, and padding or fill costs. Figure 2 shows these effects on a common matrix. They are coupled to the parallel decomposition selected by the GPU kernel.

These interactions motivate two questions. First, with the operator and GPU fixed, how strongly does the profitable implementation depend on the sparsity pattern? Second, how much of this pattern-dependent optimization space is covered by existing specialized systems and sparse compilers?

2.2. Sensitivity to Sparsity Patterns

Figure 3. Sparse kernel performance sensitivity on ten SuiteSparse matrices (Davis and Hu, 2011). Top: cuSPARSE implementations for different sparse representations in SpMV and SpMM at K=128K{=}128. Bottom: three specialized systems and one sparse compiler. The best design changes with the pattern, while mismatched choices create severe slowdowns. All measurements use an NVIDIA RTX PRO 6000 Blackwell GPU with CUDA 13.0 and report the mean of ten timed iterations after warmup.

We first hold the library and platform fixed to cuSPARSE (NVIDIA Corporation, 2026), then vary the representation and SpMM width. The top panel of Figure 3 uses format names for brevity, but every bar measures a complete implementation.

The observed spread follows the assumptions of each implementation. SELL-C wins by up to 1.47×1.47\times on a stencil with only 0.08% slice padding. It loses 33×33\times when uneven row lengths inflate storage by 30×30\times. BSR wins 1.25×1.25\times when occupied 16×1616{\times}16 tiles are 87% dense. It falls to 0.04×0.04\times on power-law graphs because the occupied blocks contain mostly zeros. COO pays for explicit row metadata and reductions on regular inputs. Under row imbalance, its nonzero-level decomposition makes it the most robust SpMM choice. COO wins 21 of 30 configurations and reaches 2.8×2.8\times. Across the full sweep, the best alternative gains at most 2.8×2.8\times over CSR. A mismatched design loses up to 350×350\times.

The operator context changes the same tradeoff. At K=32K{=}32, block-dense matrices can favor CUDA-Core blocked kernels. At K=128K{=}128, Tensor-Core layouts replace them on some inputs. Blocked-ELL accelerates dense-tile arithmetic but pads every block row to a common width. This padding can increase transferred bytes by two orders of magnitude. For SpGEMM, the corresponding choices appear in row bucketing and phase structure. The accumulator dataflow also depends on the intermediate-work distribution  (Niu et al., 2022; Nagasaka et al., 2017; Parger et al., 2020; Wu et al., 2025).

Observation 1. The profitability of a sparse optimization depends strongly on the input structure. Representation and execution schedule must therefore be coordinated rather than chosen independently. A poor match can be far more costly than the gain from a good one.

2.3. Limitations of Existing Works

We next evaluate three recent operator-specialized systems and one sparse compiler on the same NVIDIA RTX PRO 6000 GPU platform. CB-SpMV (Cong et al., 2025) combines cache-friendly blocks with adaptive per-block formats. DTC-SpMM (Fan et al., 2024) combines compressed block metadata with Tensor-Core execution. HSMU-SpGEMM (Wu et al., 2025) uses a shared-memory hash accumulator for SpGEMM. SparseTIR (Ye et al., 2023) supports dynamic format decomposition and scheduling for SpMM.

Figure 3 shows how they perform. CB-SpMV achieves 1.71.72.5×2.5\times while its blocked working set remains cache-friendly. It loses on every matrix above 3 M nonzeros and falls to 0.32×0.32\times. DTC-SpMM reaches 3.1×3.1\times on block-structured inputs. It also beats cuSPARSE’s padded Tensor-Core format by 2626106×106\times on irregular matrices. However, under extreme row imbalance DTC-SpMM falls to 0.04×0.04\times. HSMU-SpGEMM surpasses cuSPARSE on three and loses on seven of ten matrices. None of the three systems dominates the full set.

The target platform changes the useful region further. These systems were developed for earlier GPU and CUDA generations. On Blackwell and CUDA 13.0, their fixed resource choices meet different hardware conditions, and they adapt poorly to the new platform. DTC-SpMM, for example, reports more than 1.5×1.5\times over cuSPARSE on an RTX 4090 (Fan et al., 2024). Running the released artifact on our Blackwell GPU with CUDA 13.0, we measure it below cuSPARSE at every dense width, from 0.73×0.73\times at K=8K{=}8 down to 0.58×0.58\times at K=256K{=}256.

Sparse tensor compilers expose a broader space through dynamic format decomposition. They also provide scheduling primitives. SparseTIR (Ye et al., 2023) beats cuSPARSE on 25 of 30 SpMM configurations. Its geometric-mean speedup is 1.40×1.40\times and its maximum is 5.0×5.0\times. SparseTIR nevertheless falls to 0.130.130.20×0.20\times on the matrix behind the 350×350\times format cliff because long rows serialize execution. SparseTIR can express a bucketed hybrid layout, but a developer supplies the format rewrite and schedule (Ye et al., 2023). Its coverage therefore depends on the implementations encoded in its rule set. It also does not support SpMV or SpGEMM, and it covers only a limited set of widths (K=32K=32, 128128, 256256) for SpMM.

Observation 2. No fixed implementation or developer-defined optimization space consistently covers different sparsity patterns, operators, and GPU architectures. The useful optimization region changes with both the input structure and the target platform.

Together, Observations 1 and 2 show that sparse kernel optimization is both pattern dependent and architecture dependent. This motivates treating sparse kernel construction as a per-workload compilation problem. Representation, execution schedule, and hardware mapping should be constructed together for the input structure and target GPU.

3. SparseDitto Design

SparseDitto treats sparse GPU kernel construction as a per-workload compilation problem. A compilation task is specified by a sparse input XX, an operator oo, and a target GPU HH. Given X,o,H\langle X,o,H\rangle, SparseDitto constructs a specialized CUDA implementation together with the sparse data structures required by that implementation.

SparseDitto represents the main compilation decisions with a compilation plan

(2) Π=R,S,θH,R=L,I,S=P,D.\Pi=\langle R,S,\theta_{H}\rangle,\qquad R=\langle L,I\rangle,\qquad S=\langle P,D\rangle.

The representation RR consists of a data layout LL and sparse metadata II. The layout specifies how sparse values are organized and stored, while the metadata supports traversal of the resulting representation. The execution schedule SS consists of a parallel decomposition PP and a dataflow DD. The decomposition maps work to Cooperative Thread Arrays (CTAs), warps, and threads, while the dataflow determines operand movement and accumulation. Finally, θH\theta_{H} contains target-specific choices such as tile sizes, launch configuration, and on-chip resource allocation. These decisions are coupled. A change in any of them can impact others.

Figure 4 shows the overall compilation process. SparseDitto first analyzes the sparse input and estimates the overhead. A energy-based model then ranks reusable optimization templates. The architecture-aware synthesis stage combines this ranking with the input analysis and target-GPU constraints to construct candidate compilation plans. Finally, an agentic lowering stage realizes each plan as CUDA. Sanitizer, numerical validation, and target-GPU profiling provide feedback for subsequent plan refinement.

Refer to caption
Figure 4. Overview of SparseDitto. Sparse pattern analysis and a learned cost model identify promising optimization templates. Architecture-aware synthesis constructs compilation plans that coordinate representation, execution schedule, and hardware mapping. Agentic lowering and profiling realizes these plans as CUDA code.

3.1. Structural Features and Overhead Modeling

SparseDitto performs a lightweight sparse analysis before compilation-plan synthesis. The analysis is guided by known sparse optimization mechanisms. For each mechanism, we identify the structural conditions that make it effective and estimate the overhead it can incurred, e.g., extra storage. The analysis produces three groups of continuous features:

(3) Φ(X,o)=Φintr,Φrepr,Φop.\Phi(X,o)=\left\langle\Phi_{\mathrm{intr}},\Phi_{\mathrm{repr}},\Phi_{\mathrm{op}}\right\rangle.

Φintr\Phi_{\mathrm{intr}} describes the input sparsity pattern. Φrepr\Phi_{\mathrm{repr}} estimates the overhead introduced by candidate sparse representations. Φop\Phi_{\mathrm{op}} captures operator-dependent work that is not visible from the input structure alone. Together, these groups contain 36 continuous features.

Intrinsic pattern features. Let ri=nnz(X[i,:])r_{i}=\operatorname{nnz}(X[i,:]) denote the number of nonzeros in row ii. Matrix size and density characterize the overall sparse working set. SparseDitto records the mean, maximum, and coefficient of variation (CV) of rir_{i}. It also records row-length quartiles and the fraction of empty rows. These statistics expose workload imbalance relevant to CSR5, merge-based decomposition, and SELL-CC-σ\sigma (Liu and Vinter, 2015; Merrill and Garland, 2016; Kreutzer et al., 2014).

SparseDitto also measures structure along the column dimension. Column-degree statistics capture reuse and asymmetry between rows and columns. Row span and normalized row span characterize access locality. We additionally record the mean gap between sorted column indices and the fraction of diagonal entries. To capture block structure, SparseDitto measures the coverage of occupied 16×1616\times 16 tiles and the fraction of nonzeros contained in them. This group contains 18 continuous features. SparseDitto also assigns the row-length distribution to one of five shape classes: uniform, moderate, skewed, power-law, or bimodal.

Representation-induced overhead features. The representation R=L,IR=\langle L,I\rangle changes both the working set size and the metadata traversed by a kernel. SparseDitto computes value-slot expansion for padded and blocked layouts:

ρELL\displaystyle\rho_{\mathrm{ELL}} =mmaxiri/nnz(X),\displaystyle={m\max_{i}r_{i}}/{\operatorname{nnz}(X)},
ρSELL(c)\displaystyle\rho_{\mathrm{SELL}}(c) =s|s|maxisri/nnz(X),\displaystyle={\textstyle\sum_{s}|s|\max_{i\in s}r_{i}}/{\operatorname{nnz}(X)},
(4) ρBSR(b)\displaystyle\rho_{\mathrm{BSR}}(b) =b2nnzbb/nnz(X),\displaystyle={b^{2}\operatorname{nnzb}_{b}}/{\operatorname{nnz}(X)},

where ss denotes a row slice and nnzbb\operatorname{nnzb}_{b} counts occupied b×bb{\times}b blocks. These factors estimate padded computation and value traffic. The storage estimate also accounts for slice offsets and block-column indices. Row permutations are included when required. SparseDitto evaluates ELL and three SELL slice heights. It also evaluates BSR with b{4,8,16}b\in\{4,8,16\}. These choices produce eight continuous representation features. Their values are recomputed when LL, II, or the format parameters change during plan refinement.

Operator-induced features. Input statistics alone do not expose the intermediate iteration space of SpGEMM. From Equation 1, SparseDitto computes the total, maximum, p95, and CV of uB(i)u_{B}(i) for the requested product. It also computes the total and CV for B=XTB=X^{T}. The accumulator-fit statistic

(5) Fh=1mi𝟏[uB(i)h]F_{h}=\frac{1}{m}\sum_{i}\mathbf{1}[u_{B}(i)\leq h]

reports the fraction of rows below capacities h=512h=512, 4K, and 32K. These thresholds distinguish accumulator regimes that a mean would hide. SparseDitto also computes

(6) γB=iuB(i)/nnz(C),\gamma_{B}={\textstyle\sum_{i}u_{B}(i)}/{\operatorname{nnz}(C)},

which measures the intermediate products merged into each output nonzero. The estimator uses stratified sampling inspired by dataflow analysis in (Srivastava et al., 2020). The ten features in this group capture intermediate work, imbalance, accumulator demand, and output compression.

Hardware and task context. Target-GPU properties are kept separate from the 36 sparse features. SparseDitto records the number of SMs and physical cache capacities in bytes. It also records per-CTA hardware limits on threads, registers, and shared-memory size. The operator and the dense width KK for SpMM specify the task context. These values are used by the learned cost model when applicable and by the architecture-aware synthesis stage to construct and check compilation plans.

3.2. Energy-Based Model for Optimization Templates

The learned cost model provides a prior over reusable optimization templates. We adopt an additive parameterization so that each ranking can be decomposed into feature-level contributions and passed to compilation-plan synthesis. It uses measured experience to rank template set VoV_{o}, that has a large effect on representation overhead, workload decomposition, or dataflow (Sedaghati et al., 2015; Zhao et al., 2018; Fan et al., 2024; Wu et al., 2025; Ye et al., 2023; Cong et al., 2025; NVIDIA Corporation, 2026).

For SpMV, VoV_{o} contains CSR, COO, SELL with slice heights {16,32,512}\{16,32,512\}, BSR with block sizes {4,8,16}\{4,8,16\}, and adaptive 2D blocking. The SpMM set contains CSR, COO, the same three BSR variants, Blocked-ELL-{16,32}\{16,32\}, and the DTC Tensor-Core design (Fan et al., 2024). The three SpGEMM templates are CSR-hash, shared-memory tiled hash, and hybrid accumulation. These templates draw on established GPU kernels (Kreutzer et al., 2014; Srivastava et al., 2020).

The learned cost model is a single additive energy model (LeCun et al., 2006; Hastie and Tibshirani, 1990). Its input xx concatenates the 36 sparse features, a one-hot shape class, an operator indicator, and the tile width log2K\log_{2}K for SpMM. For each scalar coordinate xjx_{j}, a nonlinear basis shared across operators produces

(7) hj(xj)=tanh(xjw1j+b1j)16.h_{j}(x_{j})=\tanh(x_{j}w_{1j}+b_{1j})\in\mathbb{R}^{16}.

Each operator oo then has its own readout for every template sVos\in V_{o}:

(8) Eo(sx)=βo,s+jhj(xj)𝖳w2,o,j,s.E_{o}(s\mid x)=\beta_{o,s}+\sum_{j}h_{j}(x_{j})^{\mathsf{T}}w_{2,o,j,s}.

Lower energy indicates a stronger recommendation. Sharing the basis across operators lets a single feature response be estimated from all three tasks, while the operator-specific readouts capture the different effect of that feature on SpMV, SpMM, and SpGEMM. The nonlinear basis can express threshold behavior, such as a sharp penalty once SELL padding exceeds a level, without a hand-designed cutoff.

We train the energy directly from measured runtimes. For a training example, let MVoM\subseteq V_{o} be the templates actually benchmarked and ts=log2(baseline runtime/runtime of s)t_{s}=\log_{2}(\text{baseline runtime}/\text{runtime of }s) their measured log-speedups. The objective combines a listwise likelihood (Cao et al., 2007) with a gap-weighted pairwise margin (Burges et al., 2005):

(9) o=(x,𝐭)[\displaystyle\mathcal{L}_{o}=\sum_{(x,\mathbf{t})}\bigg[ sMqslogpo(sx)\displaystyle-\sum_{s\in M}q_{s}\log p_{o}(s\mid x)
+λos,sMts>ts(tsts)softplus(Eo(sx)Eo(sx))]\displaystyle+\lambda_{o}\!\!\sum_{\begin{subarray}{c}s,s^{\prime}\in M\\ t_{s}>t_{s^{\prime}}\end{subarray}}\!\!(t_{s}-t_{s^{\prime}})\,\operatorname{softplus}\!\big(E_{o}(s\mid x)-E_{o}(s^{\prime}\mid x)\big)\bigg]

where po(sx)=softmaxsM(Eo(sx))p_{o}(s\mid x)=\mathrm{softmax}_{s\in M}\!\big(-E_{o}(s\mid x)\big) and qs=softmaxsM(ts/T)q_{s}=\mathrm{softmax}_{s\in M}\!\big(t_{s}/T\big) are soft targets from the measured speedups. The listwise term keeps near-optimal alternatives ranked highly, while the margin term orders templates by measured speed with a penalty proportional to the speedup gap. The operator-specific weight λo\lambda_{o} controls the strength of the margin term. It is smaller for SpMV, whose template speedups cluster more tightly, and larger for SpMM and SpGEMM.

The additive form decomposes each score into per-feature terms hj(xj)𝖳w2,o,j,sh_{j}(x_{j})^{\mathsf{T}}w_{2,o,j,s}, so the model is interpretable by construction (Agarwal et al., 2021). The synthesis stage receives these terms as feature-level explanations for the template ranking. The model returns a ranked set of templates. The ranking therefore guides the known part of the optimization space.

Refer to caption
Figure 5. Architecture-aware synthesis of one SpGEMM compilation plan for Shen/e40r0100 on an RTX PRO 6000.

3.3. Architecture-Aware Compilation-Plan Synthesis

The synthesis stage combines three inputs: the sparse-analysis results, the learned template ranking, and the target-GPU specification, as shown in Figure 4. It expands promising templates into complete compilation plans. Each plan records RR, SS, and θH\theta_{H}, together with storage estimates, representation overhead modeling, and on-chip resource bounds.

Workload-scale analysis. SparseDitto first estimates the storage and work exposed by a candidate design. Let

(10) Mrep(Π)=Mvalue(L)+Mmeta(I)(Bytes)M_{\mathrm{rep}}(\Pi)=M_{\mathrm{value}}(L)+M_{\mathrm{meta}}(I)\qquad(\mathrm{Bytes})

be the representation footprint. MvalueM_{\mathrm{value}} multiplies the number of stored value slots by the bytes per value. MmetaM_{\mathrm{meta}} similarly accounts for the byte width of each metadata field. Let W(Π)W(\Pi) denote temporary storage introduced by the dataflow, also in bytes. The read-side and total working sets are

(11) MR(Π)\displaystyle M_{R}(\Pi) =Mrep(Π)+W(Π)+DR\displaystyle=M_{\mathrm{rep}}(\Pi)+W(\Pi)+D_{R} (Bytes)\displaystyle(\mathrm{Bytes})
(12) MT(Π)\displaystyle M_{T}(\Pi) =MR(Π)+DS\displaystyle=M_{R}(\Pi)+D_{S} (Bytes)\displaystyle(\mathrm{Bytes})

where DRD_{R} and DSD_{S} are the byte footprints of common input and output operands. For SpMM, these terms include dense BB and CC. For SpGEMM, W(Π)W(\Pi) includes the selected accumulator and intermediate representation.

Let CL2C_{\mathrm{L2}} denote the physical L2 capacity in bytes. The scale ratios are

(13) qR(Π)\displaystyle q_{R}(\Pi) =MR(Π)CL2,\displaystyle=\frac{M_{R}(\Pi)}{C_{\mathrm{L2}}},
qT(Π)\displaystyle q_{T}(\Pi) =MT(Π)CL2.\displaystyle=\frac{M_{T}(\Pi)}{C_{\mathrm{L2}}}.

SparseDitto also reports the number of independent work units per SM, Nwork(Π)/NSMN_{\mathrm{work}}(\Pi)/N_{\mathrm{SM}}. A work unit can be a row, slice, nonzero partition, or tile. These quantities are static scale indicators. They do not predict cache residency or achieved parallelism. Both depend on the lowered code and are determined through target-GPU measurements.

Joint representation-schedule synthesis. The second stage uses these estimates and the learned template ranking to construct RR and SS together. It selects a sparse algorithm and specifies the layout, traversal metadata, parallel decomposition, and dataflow needed to realize it. For example, a SELL plan couples row slicing and slice offsets with a schedule that processes slices in parallel. A SpGEMM hash plan couples row bucketing with the accumulator dataflow. These choices determine both the work assigned to each computing unit and the temporary storage it requires.

Hardware mapping. The third stage binds RR and SS to target-specific parameters in θH\theta_{H}, including tile sizes, launch geometry, and on-chip resource allocation. For a hardware mapping tt, let BtB_{t} be its threads per CTA, rtr_{t} its registers per thread, and StS_{t} its shared-memory bytes per CTA. Let TmaxT_{\max}, RmaxR_{\max}, and SmaxS_{\max} denote the corresponding hardware limits. A legal mapping must satisfy

(14) BtTmax,rtBtRmax,StSmax.B_{t}\leq T_{\max},\qquad r_{t}B_{t}\leq R_{\max},\qquad S_{t}\leq S_{\max}.

SparseDitto can generate multiple different compilation plan candidates to expand its coverage of the design space. Each completed plan starts an independent search branch. Initial 5 iterations preserve optimization diversity, while later iterations favor the fastest valid plans. Before lowering, the synthesis stage checks the requested thread count and shared-memory allocation. After compilation with nvcc, the reported register count completes the check.

Figure 5 illustrates the synthesis of one SpGEMM plan for Shen/e40r0100 on an RTX PRO 6000. Here, C=A2C=A^{2}, m=n=17,281m=n=17{,}281, and nnz(A)=553,562\operatorname{nnz}(A)=553{,}562. Scanning the CSR structure gives maxiri=62\max_{i}r_{i}=62 and nnzb16=20,130\operatorname{nnzb}_{16}=20{,}130. Substituting these counts into Equation 4 gives ρELL=17,281×62/553,5621.9\rho_{\mathrm{ELL}}=17{,}281\times 62/553{,}562\approx 1.9 and ρBSR(16)=162×20,130/553,5629.3\rho_{\mathrm{BSR}}(16)=16^{2}\times 20{,}130/553{,}562\approx 9.3. The same block count gives 553,562/20,13027.5553{,}562/20{,}130\approx 27.5 nonzeros per occupied tile, below the 32 threads in a warp; only 20,130/17,281/1621.7%20{,}130/\lceil 17{,}281/16\rceil^{2}\approx 1.7\% of all tiles are occupied. The two histograms show rir_{i} and uA(i)=kA[i,:]rku_{A}(i)=\sum_{k\in A[i,:]}r_{k} from Equation 1; the latter ranges from 102 to 2,187 intermediate products per row.

With FP32 values and 32-bit indices, Equation 10 gives the CSR input footprint as 8×553,562+4×(17,281+1)=4,497,6248\times 553{,}562+4\times(17{,}281+1)=4{,}497{,}624 bytes, or 4.29MB. This explains the figure’s comparison with the GPU’s 128 MB L2 capacity. The complete plan’s working sets and scale ratios additionally include its workspace and other operands through Equations 12 and 13.

These estimates guide a packed column-window design that preserves diagonal locality while avoiding dense-block padding. The 4,096-column window and 256-thread CTA are synthesis choices in θH\theta_{H}. The window width yields 5 output-column strips in RR, and SS assigns one CTA to each row–window pair.

3.4. Agentic Lowering and Refinement

Each compilation plan is realized as executable CUDA through an agentic lowering stage. The lowering stage receives the selected representation, execution schedule, and hardware mapping from Π\Pi. Its role is to realize these decisions using target-specific CUDA code.

SparseDitto provides conversion tools and metadata builders for CSR, COO, ELL, SELL, and BSR. It also provides adaptive 16×1616{\times}16 blocks inspired by CB-SpMV (Cong et al., 2025). Packed column windows for SpMM follow Sextans (Song et al., 2022). SpGEMM support includes upper-bound analysis and row binning. Scans and workspace management follow the row-wise structure of MatRaptor (Srivastava et al., 2020). These mechanisms implement LL and II while providing building blocks for PP and DD.

An LLM-based coding agent realizes the thread mapping, accumulator design, and θH\theta_{H} parameters specified by the plan within a unified harness. Every generated implementation is compiled and numerically validated. SparseDitto then measures latency and profiles individual CUDA kernels with NCU on the target GPU. Profiling feedback is mapped back to the corresponding fields of Π\Pi. Excessive padding or metadata traffic triggers changes to LL or II. Imbalance and synchronization overhead trigger changes to PP. Cache behavior and workspace traffic guide changes to DD. Hardware resource pressure changes θH\theta_{H}. The synthesis stage updates the compilation plan before the coding agent revises the code.

SparseDitto returns the fastest numerically valid implementation across all branches. The cached result is identified by its sparsity pattern, operator, and target platform. For SpMM, the key also records KK. Subsequent invocations with the same key execute the compiled implementation without repeating plan synthesis or lowering.

4. Evaluation

Workloads. We use a stratified set of 6060 matrices from the SuiteSparse Matrix Collection (Davis and Hu, 2011). It covers regular stencils, finite-element matrices, circuit and process-simulation problems, power-law web graphs, road networks, and gene networks, and includes the hardest cases from the evaluations of the specialized systems we compare against. For each matrix, we evaluate SpMV, SpMM with K{8,32,128,256}K\in\{8,32,128,256\}, and SpGEMM. For SpGEMM, we follow (Wu et al., 2025), computing AAAA for square matrices and AA𝖳AA^{\mathsf{T}} otherwise.

The matrices have 359–16,777,216 rows and 817 to 101 million nonzeros. Their densities range from 1.78×1071.78\times 10^{-7} to 0.110. Mean row length ranges from 1.0 to 1,107.1 nonzeros, and the coefficient of variation of row length ranges from 0 to 12.2. Appendix C lists every matrix with these statistics.

Platform and baselines. We evaluate on two GPU architectures, NVIDIA RTX PRO 6000 Blackwell with CUDA 13.0 and NVIDIA H200 Hopper with CUDA 12.9. We report the main results on both. The ablation study and the cache and memory-traffic analysis run on the RTX PRO 6000. We use the corresponding cuSPARSE implementation (NVIDIA Corporation, 2026) as the per-task reference. We benchmark all applicable cuSPARSE algorithms under the same numerical contract and use the fastest valid configuration as the baseline. For SpMV, we compare with CB-SpMV (Cong et al., 2025) and AlphaSparse (Du et al., 2022), a auto-tuner that creates machine-designed SpMV formats and kernels. For SpMM, we compare with DTC-SpMM (Fan et al., 2024) and SparseTIR (Ye et al., 2023), a sparse compiler for SpMM. For SpGEMM, we compare with HSMU-SpGEMM (Wu et al., 2025).

LLM configuration. The agents used for plan synthesis, CUDA lowering, and verification call the same model: GPT-5.6-terra (OpenAI, 2026) with reasoning effort low and a budget of 8,1928{,}192 tokens per call. All other parameters are left at the API default. Every task uses the same search budget. Three candidate plans start independent branches, each with up to five lowering-and-evaluation iterations. The two fastest valid branches receive up to five additional iterations for profile-guided refinement.

Measurement protocol. We time the generated kernel and baselines with CUDA events. Each implementation has five warmup executions followed by ten timed executions, and we report their arithmetic mean, excluding the one-time preprocessing. Aggregate speedups are geometric means over tasks, reported with a 95%95\% confidence interval obtained by bootstrapping the tasks (20,00020{,}000 resamples of the log-speedups).

Correctness validation. We use the corresponding FP32 cuSPARSE implementation as per-task correctness reference. SparseDitto and cuSPARSE receive identical input values. For SpMV and SpMM, a candidate is considered correct only if every output entry satisfies

(15) |y^iyiref|ϵabs+ϵrel|yiref|,\left|\widehat{y}_{i}-y_{i}^{\mathrm{ref}}\right|\leq\epsilon_{\mathrm{abs}}+\epsilon_{\mathrm{rel}}\left|y_{i}^{\mathrm{ref}}\right|,

where ϵabs=105\epsilon_{\mathrm{abs}}=10^{-5} and ϵrel=103\epsilon_{\mathrm{rel}}=10^{-3}. Any runtime error, unexpected NaN or infinity, or violation of this bound causes the candidate to fail validation.

For SpGEMM, correctness includes both the sparse output structure and the associated values. We canonicalize the SparseDitto and cuSPARSE outputs by sorting column indices within each row and coalescing duplicate entries. The output dimensions, number of nonzeros, row pointers, and column indices must match exactly. Every corresponding nonzero value must additionally satisfy the same mixed absolute and relative error bound above. We apply this single predeclared validation rule uniformly to all matrices, generated candidates, and comparison systems.

Figure 6. Generated kernel performance for SpMV, SpMM, and SpGEMM. (a) geometric-mean speedup over cuSPARSE; whiskers are 95%95\% bootstrap confidence intervals over tasks. (b) the per-matrix distribution on a logarithmic scale. The dashed line is cuSPARSE.
Figure 7. The same measurement on the H200. Both panels follow Figure 6.

4.1. Generated Kernel Performance

We organize the results by operator and report each one on both GPUs. On seven of the largest matrices, the cuSPARSE SpGEMM implementation triggers an out-of-resources error, so those tasks have no reference and are excluded. Each additional system is evaluated only for the operator and configurations supported by its released artifact. Aggregate speedups are geometric means over the validated tasks.

On the RTX PRO 6000, SparseDitto reaches 2.68×2.68\times over cuSPARSE (95%95\% CI [2.48,2.89][2.48,2.89]). On the H200, the same compilation framework reads the new target’s device properties and regenerates every implementation from scratch. It reaches 2.79×2.79\times (95%95\% CI [2.61,2.99][2.61,2.99]). Figures 6 and 7 break down the results by operator and dense width.

SpMV. Against cuSPARSE, SparseDitto achieves 2.53×2.53\times on the RTX PRO 6000 and 2.68×2.68\times on the H200. CB-SpMV achieves 1.35×1.35\times and 1.70×1.70\times on the two GPUs. On the same inputs, SparseDitto is 1.91×1.91\times and 1.57×1.57\times faster than CB-SpMV.

AlphaSparse generates a customized SpMV implementation for each matrix by searching a design space defined by static rules. With a nine-hour search budget per matrix, it produces a kernel that passes our validation rule on 13 matrices. The remaining runs crash or return values outside the error bound. On those 13 matrices, SparseDitto is 5.34×5.34\times faster than AlphaSparse.

SpMM. Against cuSPARSE, SparseDitto achieves 4.46×4.46\times at K=8K=8, 2.81×2.81\times at K=32K=32, 1.92×1.92\times at K=128K=128, and 1.68×1.68\times at K=256K=256. The gain is largest at the small dense widths and narrows as KK grows. The H200 follows the same trend, at 4.18×4.18\times, 2.81×2.81\times, 1.96×1.96\times, and 1.87×1.87\times.

For SpMM, we also compare with SparseTIR and DTC-SpMM. SparseTIR reaches 1.68×1.68\times at K=32K=32. Its speedup is 1.18×1.18\times at both K=128K=128 and K=256K=256. DTC-SpMM achieves 0.73×0.73\times at K=8K=8 and 0.62×0.62\times, 0.58×0.58\times, and 0.58×0.58\times at the three larger widths, so it stays below cuSPARSE at every width. Across the evaluated tasks, SparseDitto is 1.57×1.57\times faster than SparseTIR and 3.99×3.99\times faster than DTC-SpMM. On the H200, the corresponding speedups are 1.63×1.63\times and 4.32×4.32\times.

SpGEMM. SparseDitto achieves 3.70×3.70\times over cuSPARSE on the RTX PRO 6000 and 4.09×4.09\times on the H200 (up to 146.61×146.61\times). HSMU-SpGEMM achieves 0.74×0.74\times on the square matrices, and 0.32×0.32\times on the H200. On these matched inputs, SparseDitto is 5.37×5.37\times faster than HSMU-SpGEMM on the RTX PRO 6000 and 13.60×13.60\times faster on the H200. SparseDitto also covers the rectangular inputs that HSMU-SpGEMM does not support.

Panel (b) of both figures reports the per-matrix distribution. At K=8K=8 and K=32K=32, the whole SpMM distribution sits above cuSPARSE on both GPUs. As KK grows, the box contracts toward the baseline. On the RTX PRO 6000, the interquartile range narrows from 3.283.285.57×5.57\times at K=8K=8 to 1.271.272.22×2.22\times at K=256K=256. SpGEMM has the widest spread on both platforms, with maxima of 146.61×146.61\times and 78.5×78.5\times. This variation is consistent with its sensitivity to intermediate-product distributions and accumulator design, beyond the input nonzero count alone.

4.2. Cache and Memory-Traffic Analysis

We use NVIDIA Nsight Compute to examine how the generated implementations use the memory hierarchy and GPU resources. We profile the timed kernels of each winning implementation, its cuSPARSE reference, and the specialized systems on the same inputs. The profiler replays each launch with flushed caches and locked base clocks. When a compute phase launches several kernels, we pool the sector counters over all launches. DRAM traffic is measured in bytes read and written at device memory. Table 1 summarizes cache hit rates and request counts, together with achieved occupancy and DRAM bandwidth.

Table 1. SparseDitto against each system on evaluated tasks. Δ\DeltaL1 and Δ\DeltaL2 are our hit rate minus theirs, in percentage points. The remaining columns are ours divided by theirs: L1 and L2 sector requests per operator invocation, achieved occupancy, and achieved DRAM bandwidth.
Comparison Δ\DeltaL1 L1 req. Δ\DeltaL2 L2 req. Occ. BW
vs. cuSPARSE +4.7+4.7 1.09×1.09\times +3.6+3.6 0.77×0.77\times 1.96×1.96\times 1.98×1.98\times
vs. CB-SpMV +1.6+1.6 0.63×0.63\times 7.3-7.3 0.57×0.57\times 0.93×0.93\times 1.03×1.03\times
vs. SparseTIR 18.6-18.6 0.85×0.85\times 2.2-2.2 1.03×1.03\times 1.24×1.24\times 1.80×1.80\times
vs. DTC-SpMM 5.1-5.1 1.44×1.44\times 1.3-1.3 1.13×1.13\times 3.21×3.21\times 5.03×5.03\times
vs. HSMU-SpGEMM 16.3-16.3 0.83×0.83\times +1.5+1.5 0.97×0.97\times 0.88×0.88\times 4.25×4.25\times

Versus cuSPARSE. SparseDitto issues slightly more L1 requests but fewer L2 requests, with higher hit rates at both levels. Its achieved occupancy and DRAM bandwidth are both about twice those of cuSPARSE. The gains thus accompany improvements in both memory access behavior and hardware utilization. The cache effects differ by operator. For SpMM, the median L1 hit-rate increase ranges from 88 to 1717 percentage points across dense widths. For SpGEMM, the improvement appears primarily in L2.

Versus specialized systems. SparseTIR and HSMU-SpGEMM have higher L1 hit rates than SparseDitto, yet SparseDitto is 1.57×1.57\times and 5.37×5.37\times faster on the same inputs. Against SparseTIR, SparseDitto issues 0.85×0.85\times as many L1 requests and 1.03×1.03\times as many L2 requests, while sustaining 1.80×1.80\times the DRAM bandwidth. Against CB-SpMV, it reduces L1 and L2 requests to 0.63×0.63\times and 0.57×0.57\times, despite a lower L2 hit rate.

Against HSMU-SpGEMM, SparseDitto achieves 0.88×0.88\times the occupancy but 4.25×4.25\times the DRAM bandwidth. The profiled execution uses 55 kernels per SpGEMM, compared with 2424 for HSMU-SpGEMM. Fewer stages reduce launch overhead and the number of inter-kernel dependencies. This comparison highlights the importance of the accumulation dataflow and kernel decomposition, beyond cache hit rates or occupancy.

DTC-SpMM illustrates the interaction between representation overhead and hardware mapping. It moves slightly less data than SparseDitto, yet SparseDitto is 3.99×3.99\times faster. SparseDitto achieves 3.21×3.21\times its occupancy and 5.03×5.03\times its DRAM bandwidth. DTC-SpMM’s Tensor-Core pipeline reaches only 1.5%1.5\% of peak, and its 16×816\times 8 tiles have a median nonzero occupancy of 10.1%10.1\% on these matrices. This low tile occupancy leaves much of the blocked computation operating on padding. Low achieved warp occupancy further limits latency hiding. SparseDitto’s advantage decreases from 6.40×6.40\times at K=8K=8 to 2.89×2.89\times at K=256K=256 as the dense width grows.

Together, these comparions reveal that higher cache hit rates or occupancy alone do not gurantee better performance. SparseDitto outperms baselines by covering broader optimization space insead of focusing on any single metrics.

4.3. Ablation Study and Comparison with General Agentic CUDA Coding Framework

We firstly compare with one SOTA general agentic CUDA coding framework StitchCUDA (Li et al., 2026), whose workflow and prompts design are tailored for general CUDA tasks. It achieves 1.53×1.53\times, 1.76×1.76\times, and 0.86×0.86\times over cuSPARSE on three operators seperately. Then we evaluate three configurations with progressively richer sparse-specific guidance. Learned Model + Agentic Lowering generates implementations with guide from our learned cost model. The third augments the task with static pattern analysis as input to agentic lowering stage. The full configuration adds architecture-aware synthesis of compilation plans that coordinate representation, schedule, and hardware mapping.

Table 2 reports geometric-mean speedups for each configuration. Learned Model + Agentic Lowering already outperforms cuSPARSE on average. Pattern Analysis raises SpMV speedup from 1.93×1.93\times to 2.27×2.27\times and SpMM speedup from 2.02×2.02\times to 2.28×2.28\times. The larger gain is on SpGEMM, from 1.22×1.22\times to 2.24×2.24\times. The full configuration further improves all three operators, reaching 2.53×2.53\times, 2.52×2.52\times, and 3.70×3.70\times, respectively. This progression supports sparse-specific guidance and architecture-aware plan synthesis significantly contribute to generated kernels’ performance.

Table 2. Ablation results by operator. Each cell reports the geometric-mean speedup over cuSPARSE.
Configuration SpMV SpMM SpGEMM
StitchCUDA (Li et al., 2026) 1.53×1.53\times 1.76×1.76\times 0.86×0.86\times
Learned Model + Agentic Lowering 1.93×1.93\times 2.02×2.02\times 1.22×1.22\times
+ Pattern Analysis 2.27×2.27\times 2.28×2.28\times 2.24×2.24\times
+ Architecture-aware Synthesis (full) 2.53×\mathbf{2.53\times} 2.52×\mathbf{2.52\times} 3.70×\mathbf{3.70\times}

4.4. Learned Cost Model

We next evaluate the learned cost model’s ability to rank optimization templates and the structural features that influence its ranking. This ranking supplies a prior for compilation-plan synthesis; the final implementation is selected through target-GPU measurements.

Training cost. The additive energy model has 15.915.9K parameters. Training on the full corpus of 400400 matrices for 600600 epochs takes about two seconds on a single RTX PRO 6000 or H200 GPU. At inference, ranking the templates takes well under a microsecond, a negligible cost relative to plan synthesis and agentic CUDA lowering.

Ranking quality. Our corpus contains 400400 SuiteSparse matrices with measured template runtimes. It does not overlap with the evaluation set in Section 4.1. It also outperformes multilayer perceptron or gradient-boosted trees (Friedman, 2001), while retaining an additive score decomposition. Full protocols, metric definitions, and per-operator results are provided in Appendix D.

Which features matter. Figure 8 reports permutation importance, which measures the model’s sensitivity to perturbing each feature across inputs. Matrix density, nonzero count, and row count are influential across all three operators. Operator-specific features also matter: the dense width KK for SpMM and intermediate-product upper bounds for SpGEMM. These results support using both intrinsic pattern features and operator-dependent work estimates to rank templates. For an individual input, the additive score decomposition provides feature-level guidance to synthesis. Appendix B reports the complementary model-native importance measure.

Figure 8. Structural feature importance per operator. Each cell gives a feature’s share of permutation importance in percent; colors group the feature families, and a dash marks a feature the operator does not use.

4.5. End-to-End Application Evaluation

We evaluate whether SparseDitto’s specialized implementations improve application performance in a two-layer full-batch GCN on Reddit (Hamilton et al., 2017), following DTC-SpMM’s protocol (Fan et al., 2024). We use an unweighted symmetric adjacency and sweep K{8,32,128,256}K\in\{8,32,128,256\}. For each graph and dense width, SparseDitto generates the aggregation kernel once. The kernel can be reused across epochs and training runs on the target GPU. Changes to node features and model parameters do not require regenerating it. Each result averages 100100 epochs after 1010 warmup epochs. Each epoch invokes four sparse aggregations: two in the forward pass and two in the backward pass. Symmetry also lets the backward pass reuse the corresponding forward kernel. Only the sparse aggregation implementation is replaced; the rest of the training loop is unchanged.

A grouped bar chart compares epoch time for cuSPARSE, DTC-SpMM, and SparseDitto at hidden widths 8, 32, 128, and 256. SparseDitto has the lowest epoch time at every width.

Figure 9. Full-batch GCN epoch time on Reddit across hidden widths. Horizontal marks inside bars denote SpMM time. Labels above DTC-SpMM and SparseDitto report speedup over cuSPARSE.A grouped bar chart compares epoch time for cuSPARSE, DTC-SpMM, and SparseDitto at hidden widths 8, 32, 128, and 256. SparseDitto has the lowest epoch time at every width.

Figure 9 reports the result. At K=8K=8, SparseDitto reduces the average epoch time from 24.624.6 to 7.37.3 ms, yielding a 3.39×3.39\times end-to-end speedup over cuSPARSE. At K=32K=32, it reduces the epoch from 25.225.2 to 13.513.5 ms and achieves 1.87×1.87\times. The advantage narrows as the dense dimension grows, but SparseDitto remains faster than cuSPARSE. It reaches 1.01×1.01\times at K=128K=128 and 1.14×1.14\times at K=256K=256. DTC-SpMM matches cuSPARSE only at K=8K=8. Its speedups fall to 0.94×0.94\times, 0.77×0.77\times, and 0.83×0.83\times at the other widths. SparseDitto is therefore 1.31×1.31\times3.38×3.38\times faster than DTC-SpMM across the sweep.

4.6. Compilation and Lowering Cost

We report the one-time cost of constructing a specialized implementation, including plan synthesis, lowering, and target-GPU evaluation. SparseDitto takes about 3030 minutes per matrix on average under the generation budget described above. LLM inference dominates this compilation time. Each task runs at most a few dozen compile-and-measure cycles, with each cycle taking seconds to tens of seconds depending on matrix size. The API cost for the main evaluation is about $20 per matrix. These construction costs are separate from the execution times reported in Section 4.1.

AlphaSparse (Du et al., 2022) similarly treats specialized kernel generation as an offline construction task. We allow its SpMV search a nine-hour budget per matrix. On Gset/G31, a graph matrix with 99%99\% sparsity, it compiles and runs 730730 candidate kernels within that budget. Its best kernel reaches 1.35×1.35\times over cuSPARSE. SparseDitto reaches 4.46×4.46\times on the same matrix in about 3030 minutes.

5. Conclusion

The performance of sparse matrix operators on GPU depends jointly on the data representation, the execution schedule, and the hardware mapping, which need to all match the sparsity pattern. SparseDitto addresses this challenge with a unified compilation framework, covering SpMV, SpMM, and SpGEMM. Within this framework, an interpretable energy model ranks established strategy templates from structural features, an architecture-aware synthesis stage explores the design space under target-GPU constraints, and LLM agents automatically lowering this plan into CUDA code. SparseDitto outperforms cuSPARSE in evaluated tasks, achieving geometric-mean speedups of 2.68×\times on an RTX PRO 6000 and 2.79×\times on an H200. It runs 1.57×\times to 5.37×\times faster than four state-of-the-art specialized systems and accelerates full-batch GCN training by up to 3.39×\times.

References

  • Agarwal et al. (2021) R. Agarwal, L. Melnick, N. Frosst, X. Zhang, B. Lengerich, R. Caruana, and G. E. Hinton Neural additive models: interpretable machine learning with neural nets. In Advances in Neural Information Processing Systems (NeurIPS), Vol. 34, pp. 4699–4711. Cited by: §3.2.
  • Burges et al. (2005) C. Burges, T. Shaked, E. Renshaw, A. Lazier, M. Deeds, N. Hamilton, and G. Hullender Learning to rank using gradient descent. In Proc. 22nd Int. Conf. on Machine Learning (ICML), pp. 89–96. Cited by: §3.2.
  • Cao et al. (2007) Z. Cao, T. Qin, T. Liu, M. Tsai, and H. Li Learning to rank: from pairwise approach to listwise approach. In Proc. 24th Int. Conf. on Machine Learning (ICML), pp. 129–136. Cited by: §3.2.
  • Cong et al. (2025) X. Cong, F. Sun, Y. Chen, C. Xie, Y. Liu, and D. Qian CB-spmv: a data aggregating and balance algorithm for for cache-friendly block-based spmv on gpus. In Proceedings of the 39th ACM International Conference on Supercomputing, pp. 149–160. Cited by: §1, §1, §2.3, §3.2, §3.4, §4.
  • Davis and Hu (2011) T. A. Davis and Y. Hu The university of florida sparse matrix collection. ACM Transactions on Mathematical Software 38 (1), pp. 1:1–1:25. Cited by: Figure 3, §4.
  • Du et al. (2022) Z. Du, J. Li, Y. Wang, X. Li, G. Tan, and N. Sun AlphaSparse: generating high performance spmv codes directly from sparse matrices. In Proceedings of the International Conference on High Performance Computing, Networking, Storage and Analysis, SC ’22. External Links: ISBN 9784665454445 Cited by: §1, §1, §4.6, §4.
  • Fan et al. (2024) R. Fan, W. Wang, and X. Chu DTC-spmm: bridging the gap in accelerating general sparse matrix multiplication with tensor cores. In Proceedings of the 29th ACM International Conference on Architectural Support for Programming Languages and Operating Systems, Volume 3, ASPLOS ’24, New York, NY, USA, pp. 253–267. External Links: ISBN 9798400703867, Link, Document Cited by: §1, §1, §2.1, §2.3, §2.3, §3.2, §3.2, §4.5, §4.
  • Filippone et al. (2017) S. Filippone, V. Cardellini, D. Barbieri, and A. Fanfarillo Sparse matrix-vector multiplication on gpgpus. ACM Trans. Math. Softw. 43 (4). External Links: ISSN 0098-3500, Link, Document Cited by: §2.1.
  • Friedman (2001) J. H. Friedman Greedy function approximation: a gradient boosting machine. Annals of Statistics 29 (5), pp. 1189–1232. Cited by: §4.4.
  • Hamilton et al. (2017) W. L. Hamilton, R. Ying, and J. Leskovec Inductive representation learning on large graphs. In Proceedings of the 31st International Conference on Neural Information Processing Systems, NIPS’17, Red Hook, NY, USA, pp. 1025–1035. External Links: ISBN 9781510860964 Cited by: §4.5.
  • Hastie and Tibshirani (1990) T. J. Hastie and R. J. Tibshirani Generalized additive models. Chapman & Hall/CRC. Cited by: §3.2.
  • Kjolstad et al. (2017) F. Kjolstad, S. Kamil, S. Chou, D. Lugato, and S. Amarasinghe The tensor algebra compiler. Proc. ACM Program. Lang. 1 (OOPSLA). External Links: Link, Document Cited by: §1, §1.
  • Kreutzer et al. (2014) M. Kreutzer, G. Hager, G. Wellein, H. Fehske, and A. R. Bishop A unified sparse matrix data format for efficient general sparse matrix-vector multiplication on modern processors with wide simd units. 36 (5), pp. C401–C423. External Links: ISSN 1064-8275, Link, Document Cited by: §3.1, §3.2.
  • LeCun et al. (2006) Y. LeCun, S. Chopra, R. Hadsell, M. Ranzato, and F. J. Huang A tutorial on energy-based learning. In Predicting Structured Data, Cited by: §3.2.
  • Li et al. (2026) S. Li, Z. Zhang, W. Chen, Y. Luo, M. Hong, and C. Ding StitchCUDA: an automated multi-agents end-to-end gpu programming framework with rubric-based agentic reinforcement learning. In Proceedings of the 43rd International Conference on Machine Learning, ICML’26. Cited by: §1, §4.3, Table 2.
  • Liu and Vinter (2015) W. Liu and B. Vinter CSR5: an efficient storage format for cross-platform sparse matrix-vector multiplication. In Proceedings of the 29th ACM on International Conference on Supercomputing, ICS ’15, New York, NY, USA, pp. 339–350. External Links: ISBN 9781450335591, Link, Document Cited by: §1, §2.1, §3.1.
  • Merrill and Garland (2016) D. Merrill and M. Garland Merge-based parallel sparse matrix-vector multiplication. In SC ’16: Proceedings of the International Conference for High Performance Computing, Networking, Storage and Analysis, Vol. , pp. 678–689. External Links: Document Cited by: §2.1, §3.1.
  • Nagasaka et al. (2017) Y. Nagasaka, A. Nukada, and S. Matsuoka High-performance and memory-saving sparse general matrix-matrix multiplication for nvidia pascal gpu. In 2017 46th International Conference on Parallel Processing (ICPP), Vol. , pp. 101–110. External Links: Document Cited by: §2.1, §2.2.
  • Niu et al. (2022) Y. Niu, Z. Lu, H. Ji, S. Song, Z. Jin, and W. Liu TileSpGEMM: a tiled algorithm for parallel sparse general matrix-matrix multiplication on gpus. In Proceedings of the 27th ACM SIGPLAN Symposium on Principles and Practice of Parallel Programming, PPoPP ’22, New York, NY, USA, pp. 90–106. External Links: ISBN 9781450392044, Link, Document Cited by: §2.2.
  • NVIDIA Corporation (2026) NVIDIA Corporation cuSPARSE library documentation. External Links: Link Cited by: §2.2, §3.2, §4.
  • OpenAI (2026) OpenAI GPT-5.6: frontier intelligence that scales with your ambition. Note: Accessed: 2026-07-31 External Links: Link Cited by: §4.
  • Ouyang et al. (2025) A. Ouyang, S. Guo, S. Arora, A. L. Zhang, W. Hu, C. Ré, and A. Mirhoseini KernelBench: can llms write efficient gpu kernels?. External Links: 2502.10517, Link Cited by: §1.
  • Parger et al. (2020) M. Parger, M. Winter, D. Mlakar, and M. Steinberger SpECK: accelerating gpu sparse matrix-matrix multiplication through lightweight analysis. PPoPP ’20, New York, NY, USA, pp. 362–375. External Links: ISBN 9781450368186, Link, Document Cited by: §2.2.
  • Peng et al. (2024) H. Peng, X. Xie, K. Shivdikar, M. A. Hasan, J. Zhao, S. Huang, O. Khan, D. Kaeli, and C. Ding MaxK-gnn: extremely fast gpu kernel design for accelerating graph neural networks training. In Proceedings of the 29th ACM International Conference on Architectural Support for Programming Languages and Operating Systems, Volume 2, ASPLOS ’24, New York, NY, USA, pp. 683–698. External Links: ISBN 9798400703850, Link, Document Cited by: §1.
  • Sedaghati et al. (2015) N. Sedaghati, T. Mu, L. Pouchet, S. Parthasarathy, and P. Sadayappan Automatic selection of sparse matrix representation on gpus. In Proc. 29th ACM Int. Conf. on Supercomputing (ICS), Cited by: §3.2.
  • Song et al. (2022) L. Song, Y. Chi, A. Sohrabizadeh, Y. Choi, J. Lau, and J. Cong Sextans: a streaming accelerator for general-purpose sparse-matrix dense-matrix multiplication. In Proceedings of the 2022 ACM/SIGDA International Symposium on Field-Programmable Gate Arrays, FPGA ’22, New York, NY, USA, pp. 65–77. External Links: ISBN 9781450391498, Link, Document Cited by: §3.4.
  • Srivastava et al. (2020) N. Srivastava, H. Jin, J. Liu, D. Albonesi, and Z. Zhang MatRaptor: a sparse-sparse matrix multiplication accelerator based on row-wise product. In 2020 53rd Annual IEEE/ACM International Symposium on Microarchitecture (MICRO), Vol. , pp. 766–780. External Links: Document Cited by: §2.1, §3.1, §3.2, §3.4.
  • Vázquez et al. (2010) F. Vázquez, E.M. Garzón, and J.J. Fernández A matrix approach to tomographic reconstruction and its implementation on gpus. Journal of Structural Biology 170 (1), pp. 146–151. External Links: ISSN 1047-8477, Document, Link Cited by: §1.
  • Wang et al. (2023) Y. Wang, B. Feng, Z. Wang, G. Huang, and Y. Ding TC-gnn: bridging sparse gnn computation and dense tensor cores on gpus. In USENIX Annual Technical Conference (ATC), Cited by: §1, §2.1.
  • Wu et al. (2025) M. Wu, H. Luo, F. Li, Y. Zhang, Z. Tang, K. Li, J. Zhang, and C. Liu HSMU-spgemm: achieving high shared memory utilization for parallel sparse general matrix-matrix multiplication on modern gpus. In 2025 IEEE International Symposium on High Performance Computer Architecture (HPCA), pp. 1452–1466. Cited by: §1, §1, §2.1, §2.2, §2.3, §3.2, §4, §4.
  • Ye et al. (2023) Z. Ye, R. Lai, J. Shao, T. Chen, and L. Ceze SparseTIR: composable abstractions for sparse compilation in deep learning. In Proceedings of the 28th ACM International Conference on Architectural Support for Programming Languages and Operating Systems, Volume 3, ASPLOS 2023, New York, NY, USA, pp. 660–678. External Links: ISBN 9781450399180, Link, Document Cited by: §1, §1, §2.3, §2.3, §3.2, §4.
  • Zhao et al. (2018) Y. Zhao, J. Li, C. Liao, and X. Shen Bridging the gap between deep learning and sparse matrix format selection. In Proc. 23rd ACM SIGPLAN Symp. on Principles and Practice of Parallel Programming (PPoPP), pp. 94–108. Cited by: §3.2.
  • Zhu et al. (2025) X. Zhu, S. Peng, J. Guo, Y. Chen, Q. Guo, Y. Wen, H. Qin, R. Chen, Q. Zhou, K. Gao, et al. QiMeng-kernel: macro-thinking micro-coding paradigm for llm-based high-performance gpu kernel generation. arXiv preprint arXiv:2511.20100. Cited by: §1.

Appendix A Compilation-Plan Design Space

Table 3 expands the compilation-plan formulation in Section 3. It lists representative choices for the representation R=L,IR=\langle L,I\rangle, execution schedule S=P,DS=\langle P,D\rangle, and hardware mapping θH\theta_{H} of each operator. These choices are coupled: the representation constrains the parallel decomposition and dataflow, which in turn affect hardware resource use. Synthesis combines compatible choices into a complete plan. The entries describe optimization mechanisms, not a fixed set of complete kernels.

Op. Layout and metadata L,I\langle L,I\rangle Parallel decomposition PP Dataflow DD Hardware mapping θH\theta_{H}
SpMV CSR, COO, ELL, SELL-CC-σ\sigma with slice height CC and sort window σ\sigma, BSR, adaptive 16×1616\times 16 tiles with per-tile format selection. Metadata: row pointers, slice offsets, block indices, row permutations, per-tile descriptors. One thread, one warp, or one CTA per row; equal-nonzero chunks with merge-path boundaries; slice- or block-parallel; cyclic row assignment for balance. Direct write to yy; segmented reduction across chunk boundaries; atomic accumulation; read-only or texture-cached access to xx; windowed staging of xx segments in shared memory. Rows or nonzeros per CTA, slice height, block size, threads per block, active column-window width, unroll factor.
SpMM CSR, COO, BSR, Blocked-ELL, packed column windows, adaptive tiles. Metadata: row and block indices, window descriptors, Tensor-Core fragment maps. Sparse rows, nonzeros, or blocks crossed with tiles of the dense dimension KK; one warp per row window; two-level tiling over (rows, KK). Reuse of dense rows across sparse nonzeros; accumulation in registers or shared memory; BB staged in shared memory with CC held in registers; CUDA-core or Tensor-Core (mma) execution; double buffering of the sparse stream. KK tile width, block shape, warps per CTA, shared-memory staging size, fragment layout, double-buffer depth.
SpGEMM CSR, row-reordered CSR, tiled inputs, materialized A𝖳A^{\mathsf{T}} for AA𝖳AA^{\mathsf{T}}. Metadata: row bins by upper-bound work, intermediate-product counts, tile pointers. Rows grouped into bins by predicted output size; one warp or CTA per row; tile-parallel; separate symbolic and numeric phases. Dense, hash, sorted-merge, or hybrid accumulation; on-chip accumulators sized per bin; global spill for oversized rows; output compaction by prefix sum. Hash-table capacity, dense-stripe width, bin thresholds, threads per row, shared-memory allocation per CTA.
Table 3. Representative compilation-plan choices by operator. Synthesis coordinates representation, parallel decomposition, dataflow, and hardware mapping rather than selecting them independently.

Appendix B Feature Importance

Model-native importance.

The model in Equation 8 exposes each feature’s score contribution, φj,o,s(xj)=hj(xj)𝖳w2,o,j,s\varphi_{j,o,s}(x_{j})=h_{j}(x_{j})^{\mathsf{T}}w_{2,o,j,s}. Template rankings depend on differences between scores. We summarize a feature’s influence by the variation of its contribution across templates at the mean feature vector x¯\bar{x}:

(16) Ij(o)=stdsVoφj,o,s(x¯j),I_{j}(o)=\operatorname{std}_{s\in V_{o}}\;\varphi_{j,o,s}(\bar{x}_{j}),

The values are normalized to sum to one per operator. Figure 10 reports this measure for the 3636 structural features and the SpMM width KK. It complements the permutation importance in Figure 8: one describes score variation at a reference feature vector, while the other measures sensitivity to feature permutation across inputs.

Density, nonzero count, and row count are influential across operators. The dense width KK is prominent for SpMM, while intermediate-work features matter for SpGEMM. About 3030 features account for 90%90\% of the model-native importance, indicating that the scores draw on a broad set of structural signals.

A heatmap shows model-native importance for 36 structural features and the SpMM dense width. Density and matrix size contribute across operators, with additional operator-specific features.

Figure 10. Normalized model-native feature importance by operator, in percent. Features are grouped by the analysis taxonomy and ordered within each group. The dense width KK applies only to SpMM.A heatmap shows model-native importance for 36 structural features and the SpMM dense width. Density and matrix size contribute across operators, with additional operator-specific features.

Appendix C Evaluation Matrices

Table 4 lists the 6060 SuiteSparse matrices used in Section 4, sorted by nonzero count. Mean and Max denote the mean and maximum number of nonzeros per row. CV is the coefficient of variation of row length, and density is nnz(X)/(mn)\operatorname{nnz}(X)/(mn). The last column indicates whether the matrix is square.

Table 4. The 6060 evaluation matrices, sorted by nonzero count.
Matrix Rows Cols NNZ Density Mean Max CV Sq.
HB_bcsstm19 817 817 817 ×1031.2\!\times\!10^{-3} 1.0 1 0.00 yes
HB_str_0 363 363 2,454 ×1021.9\!\times\!10^{-2} 6.8 34 1.09 yes
LPnetlib_lp_standata 359 1,274 3,230 ×1037.1\!\times\!10^{-3} 9.0 745 4.41 no
HB_bp_600 822 822 4,172 ×1036.2\!\times\!10^{-3} 5.1 302 2.37 yes
Bai_cdde2 961 961 4,681 ×1035.1\!\times\!10^{-3} 4.9 5 0.07 yes
Sandia_fpga_dcop_04 1,220 1,220 5,884 ×1034.0\!\times\!10^{-3} 4.8 36 0.83 yes
Sandia_fpga_dcop_11 1,220 1,220 5,892 ×1034.0\!\times\!10^{-3} 4.8 36 0.83 yes
Sandia_fpga_dcop_47 1,220 1,220 5,892 ×1034.0\!\times\!10^{-3} 4.8 36 0.83 yes
JGD_Homology_n3c6-b10 675 2,511 7,425 ×1034.4\!\times\!10^{-3} 11.0 11 0.00 no
Oberwolfach_flowmeter0 9,669 9,669 9,669 ×1041.0\!\times\!10^{-4} 1.0 1 0.00 yes
HB_can_838 838 838 10,010 ×1021.4\!\times\!10^{-2} 11.9 32 0.52 yes
Sandia_adder_dcop_44 1,813 1,813 11,245 ×1033.4\!\times\!10^{-3} 6.2 1,310 4.96 yes
Sandia_adder_dcop_54 1,813 1,813 11,246 ×1033.4\!\times\!10^{-3} 6.2 1,310 4.96 yes
PARSEC_Si2 769 769 17,801 ×1023.0\!\times\!10^{-2} 23.1 61 0.39 yes
HB_bcsstk09 1,083 1,083 18,437 ×1021.6\!\times\!10^{-2} 17.0 23 0.24 yes
Gset_G47 1,000 1,000 19,980 ×1022.0\!\times\!10^{-2} 20.0 34 0.22 yes
JGD_G5_IG5-11 1,227 1,692 22,110 ×1021.1\!\times\!10^{-2} 18.0 111 0.83 no
Newman_hep-th 8,361 8,361 31,502 ×1044.5\!\times\!10^{-4} 3.8 50 1.14 yes
Gset_G31 2,000 2,000 39,980 ×10310.0\!\times\!10^{-3} 20.0 40 0.23 yes
JGD_Franz_Franz5 7,382 2,882 44,056 ×1032.1\!\times\!10^{-3} 6.0 6 0.04 no
DIMACS10_fe_4elt2 11,143 11,143 65,636 ×1045.3\!\times\!10^{-4} 5.9 12 0.15 yes
FIDAP_ex14 3,251 3,251 66,775 ×1036.3\!\times\!10^{-3} 20.5 37 0.38 yes
Rommes_bips07_3078 21,128 21,128 75,729 ×1041.7\!\times\!10^{-4} 3.6 27 0.76 yes
GHS_indef_aug2d 29,008 29,008 76,832 ×1059.1\!\times\!10^{-5} 2.6 4 0.36 yes
Schenk_IBMNA_c-38 8,127 8,127 77,689 ×1031.2\!\times\!10^{-3} 9.6 889 1.73 yes
Boeing_msc04515 4,515 4,515 97,707 ×1034.8\!\times\!10^{-3} 21.6 27 0.23 yes
FIDAP_ex15 6,867 6,867 98,671 ×1032.1\!\times\!10^{-3} 14.4 18 0.30 yes
Hollinger_g7jac040 11,790 11,790 114,671 ×1048.2\!\times\!10^{-4} 9.7 120 1.50 yes
Bindel_ted_B_unscaled 10,605 10,605 144,579 ×1031.3\!\times\!10^{-3} 13.6 49 0.89 yes
ML_Graph_kmnist_norm_10NN 10,000 10,000 156,932 ×1031.6\!\times\!10^{-3} 15.7 140 0.47 yes
Schenk_IBMNA_c-50 22,401 22,401 193,625 ×1043.9\!\times\!10^{-4} 8.6 1,919 2.65 yes
Rajat_rajat22 39,899 39,899 197,264 ×1041.2\!\times\!10^{-4} 4.9 3,401 5.05 yes
Toledo_deltaX 68,600 21,961 247,424 ×1041.6\!\times\!10^{-4} 3.6 83 1.63 no
JGD_Franz_Franz11 47,104 30,144 329,728 ×1042.3\!\times\!10^{-4} 7.0 7 0.00 no
JGD_GL7d_GL7d13 47,271 8,899 356,232 ×1048.5\!\times\!10^{-4} 7.5 14 0.34 no
Quaglino_viscoplastic2 32,769 32,769 381,326 ×1043.6\!\times\!10^{-4} 11.6 73 1.20 yes
Andrianov_net25 9,520 9,520 401,200 ×1034.4\!\times\!10^{-3} 42.1 139 0.77 yes
GHS_indef_c-59 41,282 41,282 480,536 ×1042.8\!\times\!10^{-4} 11.6 3,090 2.04 yes
Shen_e40r0100 17,281 17,281 553,562 ×1031.9\!\times\!10^{-3} 32.0 62 0.49 yes
JGD_Homology_ch7-8-b4 141,120 58,800 705,600 ×1058.5\!\times\!10^{-5} 5.0 5 0.00 no
AMD_G2_circuit 150,102 150,102 726,674 ×1053.2\!\times\!10^{-5} 4.8 6 0.13 yes
Watson_Baumann 112,211 112,211 760,631 ×1056.0\!\times\!10^{-5} 6.8 7 0.06 yes
QY_case39 40,216 40,216 1,042,160 ×1046.4\!\times\!10^{-4} 25.9 20,024 12.20 yes
Pereyra_landmark 71,952 2,704 1,151,232 ×1035.9\!\times\!10^{-3} 16.0 16 0.00 no
Norris_heart1 3,557 3,557 1,387,773 ×1011.1\!\times\!10^{-1} 390.2 1,120 0.32 yes
DIMACS10_ga2010 291,086 291,086 1,418,056 ×1051.7\!\times\!10^{-5} 4.9 85 0.61 yes
CEMW_vfem 93,476 93,476 1,434,636 ×1041.6\!\times\!10^{-4} 15.3 28 0.20 yes
Barabasi_NotreDame_actors 392,400 127,823 1,470,404 ×1052.9\!\times\!10^{-5} 3.7 646 2.75 no
Nemeth_nemeth26 9,506 9,506 1,511,760 ×1021.7\!\times\!10^{-2} 159.0 193 0.16 yes
JGD_Homology_ch7-9-b5 423,360 317,520 2,540,160 ×1051.9\!\times\!10^{-5} 6.0 6 0.00 no
Andrianov_net125 36,720 36,720 2,577,200 ×1031.9\!\times\!10^{-3} 70.2 231 0.95 yes
TSOPF_TSOPF_FS_b300_c2 56,814 56,814 8,767,466 ×1032.7\!\times\!10^{-3} 154.3 27,742 6.23 yes
Gupta_gupta3 16,783 16,783 9,323,427 ×1023.3\!\times\!10^{-2} 555.5 14,672 2.22 yes
LAW_in-2004 1,382,908 1,382,908 16,917,053 ×1068.8\!\times\!10^{-6} 12.2 7,753 3.04 yes
Janna_CoupCons3D 416,800 416,800 22,322,336 ×1041.3\!\times\!10^{-4} 53.6 76 0.16 yes
Belcastro_human_gene1 22,283 22,283 24,669,643 ×1025.0\!\times\!10^{-2} 1107.1 7,939 1.27 yes
DIMACS10_asia_osm 11,950,757 11,950,757 25,423,206 ×1071.8\!\times\!10^{-7} 2.1 9 0.23 yes
Belcastro_mouse_gene 45,101 45,101 28,967,291 ×1021.4\!\times\!10^{-2} 642.3 8,032 1.33 yes
Schenk_AFE_af_shell10 1,508,065 1,508,065 52,672,325 ×1052.3\!\times\!10^{-5} 34.9 35 0.03 yes
DIMACS10_delaunay_n24 16,777,216 16,777,216 100,663,202 ×1073.6\!\times\!10^{-7} 6.0 26 0.22 yes

Appendix D Learned Cost Model Evaluation

This section supplements Section 4.4 with the evaluation protocol, metric definitions, and per-operator results for the learned template-ranking model.

Protocol.

The offline corpus contains 400400 SuiteSparse matrices with measured template runtimes. Because the structural features of a matrix are shared across its tasks, we report five-fold cross-validation grouped by matrix and stratified by shape, averaged over five seeds; a separate 100100-matrix test set is held out and never used for tuning. We use four metrics per operator. Top-1 is the fraction of inputs whose top-ranked template is the measured best. Regret is the geometric-mean slowdown of the top-ranked template relative to the per-input oracle. Capture is the fraction of the oracle speedup realized by the ranking. Cliff is the fraction of inputs whose top choice is more than 2×2\times slower than the CSR baseline. Regret and capture, rather than top-1, are the quantities that reflect the performance cost of a ranking error, because selecting a near-optimal template can preserve performance despite a top-1 error.

Baselines.

We compare against a multilayer perceptron (MLP) ranker trained with the same listwise objective, and gradient-boosted trees (GBDT), a strong reference model for tabular data. All models use the same 3636 structural features and identical cross-validation folds. Energy (listwise) ablates our model by removing the margin term (λo=0\lambda_{o}=0), leaving the same architecture trained as a listwise ranker.

SpMV SpMM SpGEMM
Model top1\uparrow regret\downarrow capt.\uparrow top1\uparrow regret\downarrow capt.\uparrow top1\uparrow regret\downarrow capt.\uparrow
MLP 0.805 1.03 95% 0.715 1.17 76% 0.681 1.37 76%
GBDT 0.793 1.03 96% 0.716 1.09 87% 0.713 1.26 82%
Energy (listwise) 0.797 1.03 96% 0.716 1.14 80% 0.697 1.34 78%
Energy (ours) 0.797 1.03 96% 0.719 1.13 81% 0.702 1.32 79%
Table 5. Template-ranking quality (5-fold grouped cross-validation, averaged over 5 seeds). marks the black-box reference model. Bold marks where the energy model beats the MLP on SpMM and SpGEMM. Cliff rates are discussed in the text.

Ranking quality.

The energy model improves on the MLP in top-1 accuracy, regret, and capture for SpMM and SpGEMM (Table 5). Capture rises from 76%76\% to 81%81\% for SpMM and from 76%76\% to 79%79\% for SpGEMM. For SpMV, both models have regret 1.031.03; the energy model improves capture from 95%95\% to 96%96\%, with slightly lower top-1 accuracy.

GBDT achieves lower regret and higher capture on SpMM and SpGEMM. The additive model remains competitive in top-1 accuracy on SpMV and SpMM while exposing feature-level score contributions to synthesis. Cliff rates remain below 6%6\% for all models. The energy model’s rates are 0.0%0.0\% for SpMV and 0.5%0.5\% for SpMM, no higher than the MLP’s. Removing the margin term reduces top-1 accuracy and capture on SpMM and SpGEMM. This supports weighting pairwise rankings by measured speedup gaps.

Appendix E Case Studies

The following cases illustrate how input structure and target hardware affect the generated implementation. Table 6 reports two tasks per operator on the RTX PRO 6000 Blackwell GPU with CUDA 13.0. The SpMV and SpMM pairs contrast different row-length distributions; the SpGEMM pair contrasts intermediate work and output size. A final case compares the same SpMM task on both GPUs. Speedups use cuSPARSE unless explicitly noted.

Op. Matrix Mean Max CV Kernel (ms) Speedup
SpMV JGD_Homology_ch7-9-b5 6.0 6 0.00 0.007 2.44×2.44\times
SpMV Barabasi_NotreDame_actors 3.7 646 2.75 0.013 1.94×1.94\times
SpMM Schenk_AFE_af_shell10 (K=32K{=}32) 34.9 35 0.03 0.736 6.82×6.82\times
SpMM LAW_in-2004 (K=128K{=}128) 12.2 7,753 3.04 1.874 1.97×1.97\times
SpGEMM DIMACS10_asia_osm 2.1 9 0.23 3.771 6.45×6.45\times^{\dagger}
SpGEMM DIMACS10_delaunay_n24 6.0 26 0.22 22.009 1.19×1.19\times
Table 6. Case-study tasks. Mean and Max are nonzeros per row, CV their coefficient of variation. Speedup is over cuSPARSE, except asia_osm, where cuSPARSE SpGEMM exits with an out-of-resources error and the reference is the HSMU-SpGEMM FP64 build instead.

E.1. SpMV

JGD_Homology_ch7-9-b5: uniform row lengths.

This 423,360×317,520423{,}360\times 317{,}520 matrix has exactly six nonzeros in every row, with CV 0.000.00. ELL therefore introduces no padding. The generated plan uses column-major ELL and assigns one thread per row. Adjacent threads access the same slot in adjacent rows through consecutive addresses. The selector initially ranks SELL with slice height 3232 first. Synthesis instead produces an ELL implementation that avoids slice metadata and achieves a 2.44×2.44\times speedup over cuSPARSE.

Barabasi_NotreDame_actors: skewed rows.

The input contains 392,400392{,}400 rows and 1.471.47M nonzeros. The mean row length is 3.73.7, the maximum is 646646, and CV is 2.752.75. Padding every row to the maximum would increase storage by roughly two orders of magnitude. The generated implementation instead uses COO and partitions the nonzero stream into equal-size chunks, with one nonzero per thread. Warp-local segmented reductions combine contributions to the same row, and atomic additions merge partial sums across warps. This balances nonzero processing across warps and achieves 1.94×1.94\times. The two SpMV cases illustrate how row-length variation changes both representation and parallel decomposition.

E.2. SpMM

Schenk_AFE_af_shell10 at K=32K{=}32: regular finite-element structure.

This matrix has 1.51.5M rows and 52.752.7M nonzeros, with mean row length 34.934.9, maximum 3535, and CV 0.030.03. ELL requires little padding. The generated kernel assigns a four-thread group to each sparse row. Each thread accumulates eight output values in registers, while shuffle instructions share sparse entries within the group. The implementation pipelines sparse loads and writes each output once the row is complete. It achieves 6.82×6.82\times. Although the selector ranks a Tensor-Core block template first, measured refinement selects this CUDA-core implementation. As in Section 4.2, template rankings provide an initial prior rather than fixing the final implementation.

LAW_in-2004 at K=128K{=}128: a web graph.

This matrix has 1.41.4M rows and 16.916.9M nonzeros, with mean row length 12.212.2, maximum 7,7537{,}753, and CV 3.043.04. The generated implementation separates short rows from long rows. A warp processes each short row with register accumulation. Long rows are split into chunks of at most 256256 nonzeros, each assigned to a CTA; atomic additions combine the partial outputs. Both paths read the dense operand directly. This hybrid mapping achieves 1.97×1.97\times, with a compute time of 1.8741.874 ms. Preprocessing takes 8.98.9 s and is excluded from that measurement. Its work lists depend on the sparse structure and can be reused across subsequent calls with different dense operands. Appendix F gives the generated source.

E.3. SpGEMM

DIMACS10_asia_osm: a sparse road network.

This road network has 12.012.0M rows, 2.12.1 nonzeros per row on average, and density 1.8×1071.8\times 10^{-7}. The product AAAA has 42.842.8M nonzeros. A full-width dense accumulator would require about 1212M entries per row despite the sparse output. The generated plan instead bins rows by their intermediate-product upper bounds. Rows with bounds of at most 88 or 3636 use fixed-capacity thread-local accumulators; larger rows use a separate traversal-based path. The measured speedup is 6.45×6.45\times against the HSMU-SpGEMM FP64 build. cuSPARSE exits with an out-of-resources error on this task, so this comparison is separate from the main cuSPARSE speedup aggregate.

DIMACS10_delaunay_n24: a large sparse output.

This Delaunay triangulation has 16.816.8M rows and 100.7100.7M nonzeros, with six nonzeros per row on average. The generated plan bins rows by their intermediate-product upper bounds and assigns a warp to each row. The two bins use shared-memory hash tables with 6464 and 512512 entries, respectively. A compaction pass produces the final output, which contains 347347M nonzeros. The kernel achieves 1.19×1.19\times. Binning reduces accumulator overhead, but it does not remove the substantial traffic required to materialize this output.

E.4. The Same Task on Two GPUs

The final case fixes the matrix and varies the GPU to illustrate architecture-aware synthesis.

Nemeth_nemeth26 at K=8K{=}8 is a 9,5069{,}506-row matrix with a strongly diagonal structure. It has 159159 nonzeros per row on average with CV 0.160.16, a median row bandwidth of 185185 columns, and an average gap of 1.21.2 between consecutive column indices. Dense tiles account for 65.2%65.2\% of the nonempty 16×1616{\times}16 tiles and contain 88.2%88.2\% of the nonzeros. The structural profile and selector ranking are identical on both GPUs, with a 32×3232{\times}32 Blocked-ELL Tensor-Core template ranked first.

On the RTX PRO 6000, the selected implementation uses Tensor Cores and achieves 4.57×4.57\times over cuSPARSE. On the H200, synthesis retains a Tensor-Core candidate but also explores a SIMT block-row plan. The latter wins at 8.33×8.33\times.

The planning record illustrates how hardware capacities guide these choices. It estimates a 9.289.28 MiB working set for a 16×1616{\times}16 BSR candidate, including dense operands and output, compared with 12.1512.15 MiB for CSR. It compares these footprints with the H200’s 5050 MiB physical L2 capacity and considers an 88 KiB shared-memory tile for a windowed candidate. These estimates guide initial plans; actual cache behavior and parallelism depend on the lowered code. Target-GPU measurements then determine which implementation to retain. The different winners show why an identical structural prior need not lead to an identical implementation across GPUs.

Appendix F A Generated SpMM Implementation

Listing 1 gives the generated SpMM implementation for LAW_in-2004 at K=128K{=}128, without manual edits. It follows the harness’s preprocessing, compute, and cleanup interface. Preprocessing prepares the sparse representation and work lists. The compute entry point evaluates the operator, and cleanup releases resources. The reported 1.97×1.97\times speedup measures only the compute function, as in the main evaluation.

Preprocessing separates rows shorter than 256256 nonzeros from chunks of longer rows. The short-row kernel assigns one warp per row, with four output accumulators per thread. The long-row kernel assigns a 128128-thread CTA to each chunk and atomically combines partial outputs. Both kernels read CSR and the dense operand directly. The generated preprocessing also constructs a windowed representation, but the compute function does not use it. The executed plan combines row-wise and chunk-wise parallelism within one implementation.

Listing 1: Generated SpMM implementation for LAW_in-2004 at K=128K{=}128.
1 #include <cuda_runtime.h>
2 #include <stdint.h>
3 #include <vector>
4 #include "sparse_formats.cuh"
5
6 struct SpmmContext {
7 WindowedData windowed;
8 int* d_light_rows;
9 int light_count;
10 int* d_chunk_rows;
11 int* d_chunk_begin;
12 int* d_chunk_end;
13 int chunk_count;
14 int num_rows;
15 int nnz;
16 };
17
18 __global__ __launch_bounds__(256, 4)
19 void light_rows_spmm_kernel(
20 const int* __restrict__ row_ids,
21 int count,
22 const int* __restrict__ row_ptr,
23 const int* __restrict__ col_ind,
24 const float* __restrict__ values,
25 const float* __restrict__ B,
26 float* __restrict__ C)
27 {
28 const int tid = (int)threadIdx.x;
29 const int warp = tid >> 5;
30 const int lane = tid & 31;
31 const int idx = (int)blockIdx.x * 8 + warp;
32
33 if (idx >= count) return;
34
35 const int row = row_ids[idx];
36 const int begin = row_ptr[row];
37 const int end = row_ptr[row + 1];
38
39 float x0 = 0.0f;
40 float x1 = 0.0f;
41 float x2 = 0.0f;
42 float x3 = 0.0f;
43
44 for (int p = begin; p < end; ++p) {
45 const int col = __ldg(col_ind + p);
46 const float a = __ldg(values + p);
47 const float* bp = B + ((size_t)col << 7) + lane;
48
49 x0 = fmaf(a, __ldg(bp), x0);
50 x1 = fmaf(a, __ldg(bp + 32), x1);
51 x2 = fmaf(a, __ldg(bp + 64), x2);
52 x3 = fmaf(a, __ldg(bp + 96), x3);
53 }
54
55 float* cp = C + ((size_t)row << 7) + lane;
56 cp[0] = x0;
57 cp[32] = x1;
58 cp[64] = x2;
59 cp[96] = x3;
60 }
61
62 __global__ __launch_bounds__(128, 6)
63 void compact_dense_chunks_spmm_kernel(
64 const int* __restrict__ chunk_rows,
65 const int* __restrict__ chunk_begin,
66 const int* __restrict__ chunk_end,
67 int chunk_count,
68 const int* __restrict__ col_ind,
69 const float* __restrict__ values,
70 const float* __restrict__ B,
71 float* __restrict__ C)
72 {
73 const int task = (int)blockIdx.x;
74 if (task >= chunk_count) return;
75
76 const int lane = (int)threadIdx.x;
77 const int row = chunk_rows[task];
78 const int begin = chunk_begin[task];
79 const int end = chunk_end[task];
80
81 float s = 0.0f;
82
83 for (int p = begin; p < end; ++p) {
84 const int col = __ldg(col_ind + p);
85 const float a = __ldg(values + p);
86 s = fmaf(a, __ldg(B + ((size_t)col << 7) + lane), s);
87 }
88
89 atomicAdd(C + ((size_t)row << 7) + lane, s);
90 }
91
92 void* spmm_preprocess(
93 const int* d_row_ptr,
94 const int* d_col_ind,
95 const float* d_values,
96 int num_rows,
97 int num_cols,
98 int nnz,
99 int K)
100 {
101 (void)K;
102
103 SpmmContext* ctx = new SpmmContext;
104 ctx->d_light_rows = nullptr;
105 ctx->d_chunk_rows = nullptr;
106 ctx->d_chunk_begin = nullptr;
107 ctx->d_chunk_end = nullptr;
108 ctx->light_count = 0;
109 ctx->chunk_count = 0;
110 ctx->num_rows = num_rows;
111 ctx->nnz = nnz;
112
113 ctx->windowed = csr_to_windowed(
114 d_row_ptr, d_col_ind, d_values,
115 num_rows, num_cols, nnz, 256, 4);
116
117 if (num_rows <= 0)
118 return ctx;
119
120 std::vector<int> h_row_ptr((size_t)num_rows + 1);
121 cudaMemcpy(h_row_ptr.data(), d_row_ptr,
122 ((size_t)num_rows + 1) * sizeof(int),
123 cudaMemcpyDeviceToHost);
124
125 std::vector<int> h_light_rows;
126 std::vector<int> h_chunk_rows;
127 std::vector<int> h_chunk_begin;
128 std::vector<int> h_chunk_end;
129
130 h_light_rows.reserve(num_rows);
131
132 const int chunk_nnz = 256;
133 for (int r = 0; r < num_rows; ++r) {
134 const int begin = h_row_ptr[r];
135 const int end = h_row_ptr[r + 1];
136 const int n = end - begin;
137
138 if (n < chunk_nnz) {
139 h_light_rows.push_back(r);
140 } else {
141 for (int p = begin; p < end; p += chunk_nnz) {
142 int q = p + chunk_nnz;
143 if (q > end) q = end;
144 h_chunk_rows.push_back(r);
145 h_chunk_begin.push_back(p);
146 h_chunk_end.push_back(q);
147 }
148 }
149 }
150
151 ctx->light_count = (int)h_light_rows.size();
152 ctx->chunk_count = (int)h_chunk_rows.size();
153
154 if (ctx->light_count > 0) {
155 cudaMalloc(&ctx->d_light_rows,
156 (size_t)ctx->light_count * sizeof(int));
157 cudaMemcpy(ctx->d_light_rows, h_light_rows.data(),
158 (size_t)ctx->light_count * sizeof(int),
159 cudaMemcpyHostToDevice);
160 }
161
162 if (ctx->chunk_count > 0) {
163 const size_t bytes = (size_t)ctx->chunk_count * sizeof(int);
164 cudaMalloc(&ctx->d_chunk_rows, bytes);
165 cudaMalloc(&ctx->d_chunk_begin, bytes);
166 cudaMalloc(&ctx->d_chunk_end, bytes);
167
168 cudaMemcpy(ctx->d_chunk_rows, h_chunk_rows.data(),
169 bytes, cudaMemcpyHostToDevice);
170 cudaMemcpy(ctx->d_chunk_begin, h_chunk_begin.data(),
171 bytes, cudaMemcpyHostToDevice);
172 cudaMemcpy(ctx->d_chunk_end, h_chunk_end.data(),
173 bytes, cudaMemcpyHostToDevice);
174 }
175
176 return ctx;
177 }
178
179 void spmm_compute(
180 void* context,
181 const int* d_row_ptr,
182 const int* d_col_ind,
183 const float* d_values,
184 const float* d_B,
185 float* d_C,
186 int num_rows,
187 int num_cols,
188 int nnz,
189 int K)
190 {
191 (void)num_cols;
192 (void)nnz;
193 (void)K;
194
195 SpmmContext* ctx = static_cast<SpmmContext*>(context);
196 if (!ctx || num_rows <= 0)
197 return;
198
199 if (ctx->nnz == 0) {
200 cudaMemsetAsync(d_C, 0, (size_t)num_rows * 128 * sizeof(float), 0);
201 return;
202 }
203
204 if (ctx->chunk_count > 0) {
205 cudaMemsetAsync(d_C, 0, (size_t)num_rows * 128 * sizeof(float), 0);
206
207 if (ctx->light_count > 0) {
208 const int blocks = (ctx->light_count + 7) >> 3;
209 light_rows_spmm_kernel<<<blocks, 256>>>(
210 ctx->d_light_rows,
211 ctx->light_count,
212 d_row_ptr,
213 d_col_ind,
214 d_values,
215 d_B,
216 d_C);
217 }
218
219 compact_dense_chunks_spmm_kernel<<<ctx->chunk_count, 128>>>(
220 ctx->d_chunk_rows,
221 ctx->d_chunk_begin,
222 ctx->d_chunk_end,
223 ctx->chunk_count,
224 d_col_ind,
225 d_values,
226 d_B,
227 d_C);
228 } else if (ctx->light_count > 0) {
229 const int blocks = (ctx->light_count + 7) >> 3;
230 light_rows_spmm_kernel<<<blocks, 256>>>(
231 ctx->d_light_rows,
232 ctx->light_count,
233 d_row_ptr,
234 d_col_ind,
235 d_values,
236 d_B,
237 d_C);
238 }
239 }
240
241 void spmm_cleanup(void* context)
242 {
243 SpmmContext* ctx = static_cast<SpmmContext*>(context);
244 if (!ctx)
245 return;
246
247 if (ctx->d_light_rows)
248 cudaFree(ctx->d_light_rows);
249 if (ctx->d_chunk_rows)
250 cudaFree(ctx->d_chunk_rows);
251 if (ctx->d_chunk_begin)
252 cudaFree(ctx->d_chunk_begin);
253 if (ctx->d_chunk_end)
254 cudaFree(ctx->d_chunk_end);
255
256 free_windowed(ctx->windowed);
257 delete ctx;
258 }